# Copyright (c) 2021 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Provide threaded data archival to a File (in addition to Memory)
"""
import json
import logging
import logging.handlers
from datetime import datetime
from multiprocessing import Process, Queue
from pathlib import Path
from time import time_ns

from archivers.archive_memory import ArchiveMemory

LOG = logging.getLogger(__name__)

NS_PER_MS = 1000000
SECOND = 1000
HOUR = 60 * 60 * SECOND


class ArchiveFile(ArchiveMemory):
    """ Manage multiple readings in memory while storing them also to disk
    in a separate thread.

    Also provides a memory archiver.

    Note that the memory archiver holds _every sample_ given to it, and has
    no sense of resolution or sample buckets.  The file archiver, however,
    does partition the samples based on time taken, into bins defined by
    the resolution of the archiver.

    This means that you will get different values looking at the in-memory
    `history()` and looking at the file contents.
    """

    def __init__(self,
                 data_path,
                 root_name,
                 resolution=None,
                 null_value=None,
                 accumulate=None):
        """ Initialize the file archiver.

        File archives are stored in an hourly hierarchy:

            `<data_path>/<root_name>-<year>-<month>-<day>/H<hour>`

        Parameters
        ----------
        data_path : Path
            Where the archive is to be stored
        root_name : String
            The name of this archive (typically the device name), to be
            incorporated into the archive filename.
        resolution : Int
            The bucket size of the archive, specified in milliseconds. The
            default is to store a value for each second of operation. Do
            not make this value too small or you will consume all of the
            memory!  Make it as large as possible for your reporting.
        null_value : Any
            The value to initialize the memory buffer with.  Defaults to None.
        accumulate : Boolean
            If True, then if we get more than one sample per resolution
            period, these samples are added together. Defaults to having
            the most recent sample stored in the resolution bucker.
            Examples of `accumulate=True` would be for flow counters where
            the total flow should be accumulated; a counter example is for
            a thermometer where adding the values for a bucket wouldn't make
            sense.
        """
        self._resolution = resolution or SECOND
        self._size = HOUR // self._resolution

        # Initialize the Memory archive
        super(ArchiveFile, self).__init__(self._size)

        self._data_path = data_path
        self._root_name = root_name
        self._null_value = null_value
        self._accumulate = accumulate or False
        self._thread = None
        self._write_queue = None
        self._write_proc = None

    def run(self):
        """ Start the archive writer
        """
        if self._running:
            LOG.warning("Already running")
            return

        # We communicate to the actual file writer via a multiprocessing
        # Queue.
        self._write_queue = Queue()
        self._write_proc = Process(target=_archive_writer,
                                   args=(self._write_queue,
                                         self._data_path,
                                         self._root_name,
                                         self._resolution,
                                         self._null_value,
                                         self._accumulate))

        # Start the memory archive
        super(ArchiveFile, self).run()

    def stop(self):
        """ Stop the archive writer
        """
        if not self._running:
            LOG.warning("Already stopped")
            return

        self._write_proc.terminate()
        self._write_proc.join()

        # Stop the memory archive
        super(ArchiveFile, self).stop()

    def put(self, value):
        """ Add a value to the archive.  Also pushes the value with a
        timestamp to the write queue.
        """
        if not self._running:
            raise RuntimeError("Memory archive is not running.")

        # Store the value in the memory buffer
        super(ArchiveFile, self).put(value)

        # Send the file and timestamp to the file writer. This does mean
        # that we have two memory buffers holding this value, so be sure
        # to keep the resolution sensible.
        now = time_ns() // NS_PER_MS
        self._write_queue.put((value, now))


def _archive_writer(queue,
                    data_path,
                    root_name,
                    resolution,
                    null_value,
                    accumulate):
    """ The archive writer is a separate process that manages disk output,
    separating this slow activity from the real-time constrained file
    archiver process.

    Parameters
    ----------
    queue : Queue
        (value, timestamp) data are pushed to this queue, and received here
    data_path : Path
        Where the archive is to be stored
    root_name : String
        The name of this archive (typically the device name), to be
        incorporated into the archive filename.
    resolution : Int
        The bucket size of the archive, specified in milliseconds.
    null_value : Any
        The value to initialize the memory buffer with.
    accumulate : Boolean
        If True, then if we get more than one sample per resolution
        period, these samples are added together.
    """
    # We log to a rotating file
    LOG = logging.getLogger('archive_writer')
    handler = logging.handlers.RotatingFileHandler(
        Path(data_path, f'{root_name}_archive_writer.log'),
        maxBytes=1024 * 1024,
        backupCount=3)
    handler.setFormatter(logging.Formatter("[%(asctime)s] [%(levelname)s] "
                                           "[%(name)s.%(lineno)d] %(message)s",
                                           datefmt="%Y-%m-%d %H:%M:%S"))
    handler.setLevel(logging.INFO)
    LOG.addHandler(handler)
    LOG.setLevel(logging.INFO)

    # Create the file in-memory buffer
    size = HOUR // resolution
    buffer = [null_value] * size

    # Make our base folder
    folder = Path(data_path)
    folder.mkdir(parents=True, exist_ok=True)

    LOG.info("Starting archive writer")

    # Start up time-bin tracking
    prev_when = time_ns() // NS_PER_MS
    prev_bin = (prev_when // resolution) % size
    try:
        while True:
            # Determine current bin
            value, when = queue.get()
            bin = (when // resolution) % size

            # If we have looped our buffer, write it out to disk
            if bin < prev_bin:
                date = datetime.fromtimestamp(when // 1000)
                folder = Path(data_path,
                              f'{root_name}-'
                              f'{date.year:04}.{date.month:02}.{date.day:02}')
                folder.mkdir(parents=True, exist_ok=True)
                file = Path(folder, f'H{date.hour}')
                LOG.info(f"Writing {file}")

                # Inefficient, but saving as a json list
                data_str = json.dumps(buffer)
                try:
                    with open(file, 'w') as f:
                        f.write(data_str)
                except Exception as e:
                    LOG.error(f"ERROR WRITING:  {type(e)}: {e}")

                # Reset the buffer
                buffer = [null_value] * size

            elif bin == prev_bin:
                # Accumulate or overwrite if we are in the same bin as before
                if accumulate:
                    value = buffer[bin] + value
                buffer[bin] = value

            else:
                # bin > prev_bin; new bin!  Write to it and shift prev_bin
                buffer[bin] = value
                prev_bin = bin

    except Exception as e:
        LOG.info(f"Caught exception {type(e)}: {e}")
        LOG.info("Exiting")



# Copyright (c) 2021 Edwin Wise
# MIT License
# See LICENSE for details
"""
    PiDeviceNet Controller.  The central hub that coordinates and receives
    messages from the various Modules.
"""
import json
import logging
import socket
from dataclasses import asdict
from uuid import uuid4

import paho.mqtt.client as mqtt

from pi_device_net.constants import (
    MULTICAST_GROUP, MULTICAST_PORT, DISCOVERY_PACKET_SIZE, Command,
    ControllerPublish, ControllerDiscover, ControllerConnectReply, ModuleFound)

LOG = logging.getLogger(__name__)


class ControlNetwork():
    """ PiDeviceNet Controller

    When you initialize you specify a message callback, which is how you
    will receive all messages.

    Your application must call `controller.discover()` on a regular cycle,
    to request and process module connections.
    """

    def __init__(self,
                 system_name,
                 message_callback,
                 multicast_group=None,
                 multicast_port=None,
                 discovery_size=None):
        """ Initialize the Controller

        Parameters
        ----------
        system_name : String
            The name of the control network.  The controllers and all modules
            must be using the same system name.  You can run multiple isolated
            device networks on a single router if they have different system
            names.
        message_callback : Callable[String, String, Dict]
            For every system message received by the controller, this callable
            is called with (module_name, topic_string, message_dictionary)
        multicast_group : String
            The IP address of the multicast group used for module discovery.
            Defaults to a sensible broadcast address in `constants.py`
        multicast_port : Int
            The port that is used for the discovery socket. Defaults to the
            value in `constants.py`
        discovery_size : Int
            The size that the discovery messages are limited to.  Defaults
            to the value in `constants.py`
        """
        self.id = str(uuid4())
        self.system = system_name
        self.host = socket.getfqdn()
        self.modules = set()
        self.multicast_group = multicast_group or MULTICAST_GROUP
        self.multicast_port = multicast_port or MULTICAST_PORT
        self.discovery_size = discovery_size or DISCOVERY_PACKET_SIZE
        self._discovery_socket = self._connect_socket()
        self._mqtt = self._connect_mqtt()
        self._message_callback = message_callback

    def discover(self):
        """ Send out a discovery request and listen for any replies.

        Call this method on a regular cycle to find and process new modules.
        May take an irregular amount of time as modules are coming online.
        """

        self._broadcast_discover()
        while True:
            try:
                # See if any modules have replied
                data, address = self._discovery_socket.recvfrom(
                    self.discovery_size)
            except BlockingIOError:
                # No messages to receive, exit discovery
                break
            except Exception as e:
                LOG.error(f"Unmanaged exception {type(e)}: {e}")
                break

            # Bogus reply, so exist discovery
            if data is None or len(data) < 1:
                break

            # We expect messages to be of a particular form, transmitted
            # as json strings.
            packet = json.loads(data)
            command = packet.get('command')
            if command == Command.FOUND.value:
                found = ModuleFound(**packet)
                module = f'{self.system}/{found.module}'
                LOG.info(f"Requesting connection to {address!r}, "
                         f"module id {found.id!r}, "
                         f"as module {module!r}")
                self._reply_connect(address, found.id, module)

    def publish(self, topic_path, message):
        """ Publish a message to a given topic.

        All modules who subscribe to this topic will receive the message.

        Parameters
        ----------
        topic_path : String
            The full topic to publish, in the format:
                "<system>/<module>/<topic>"
            where:
                system = self.system
                module = module name
                topic = specific topic
        """
        cmd = ControllerPublish(system=self.system,
                                message=message)
        packet = asdict(cmd)
        payload = json.dumps(packet).encode()
        LOG.debug(f"Controller publish topic {topic_path!r}:  {payload}")
        # QOS 1 : At least once
        self._mqtt.publish(topic_path, payload, qos=1)

    def list_modules(self):
        """ List the known module topics.

        Returns
        -------
        List(String)
            The list of known modules, which are in the format:
                "<system>/<module>"
            where:
                system = self.system
                module = module name
        """
        return self.modules.copy()

    def _connect_socket(self):
        """ Connect to the discovery socket.

        Returns
        -------
        Socket
        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setblocking(False)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 20)
        return sock

    def _connect_mqtt(self):
        """ Connect to the messaging broker

        Returns
        -------
        mqtt.Client
        """
        client = mqtt.Client(client_id=self.id)
        client.on_connect = self._connect_cb
        client.on_disconnect = self._disconnect_cb
        client.on_message = self._message_cb
        client.connect(self.host)
        client.loop_start()
        return client

    def _broadcast_discover(self):
        """ Broadcast the discovery message.
        """
        cmd = ControllerDiscover(system=self.system,
                                 host=self.host)
        packet = asdict(cmd)
        self._discovery_socket.sendto(json.dumps(packet).encode(),
                                      (self.multicast_group,
                                       self.multicast_port))

    def _reply_connect(self, address, module_id, topic):
        """ Reply to a found module with a connect message.

        Parameters
        ----------
        address : (String, Int)
            Module address as returned from the socket, as a tuple
            (<host>, <port>)
        module_id : String
            The module's unique ID as reported in the found message.
        topic : String
            The topic root we are requesting the module publish to, in
            the form "<system>/<module name>"; and to subscribe to for
            administration messages, in the form "<system>/<module name>/admin"
        """
        reply = ControllerConnectReply(system=self.system,
                                       broker=self.host,
                                       id=module_id,
                                       topic=topic)
        packet = asdict(reply)
        self._discovery_socket.sendto(json.dumps(packet).encode(), address)

        # Add the topic to our topics list and subscribe to it
        self.modules.add(topic)
        all_topics = f"{topic}/+"
        LOG.info(f"Subscribing to {all_topics!r}")
        self._mqtt.subscribe(all_topics)

    def _connect_cb(self, client, userdata, flags, rc):
        """ MQTT callback on_connect; re-subscribes to all module topics.
        """
        LOG.info("MQTT Connected")
        for module in self.modules:
            all_topics = f"{module}/+"
            LOG.info(f"... subscribing to {all_topics!r}")
            client.subscribe(all_topics)

    def _disconnect_cb(self, client, topic, flags, rc=0):
        """ MQTT callback on_disconnect
        """
        LOG.warning("MQTT Disconnected")

    def _message_cb(self, client, userdata, msg):
        """ MQTT callback on_message

        Parameters
        ----------
        msg : Any
            The message is any json-parsable entity
        """
        try:
            packet = json.loads(msg.payload)
            LOG.debug(f"Controller received {packet}")
        except Exception as e:
            LOG.error(f"Unable to package message:  {type(e)}: {e}")
            return

        # Only manage messages to our system
        system = packet.get('system')
        if system != self.system:
            return

        # Only report on Module messages
        command = packet.get('command')
        if command != Command.MODULE.value:
            return

        try:
            # This system operates on three levels for the topic:
            # System name, Module name, Topic
            _, module, topic = msg.topic.split('/')
        except Exception as e:
            LOG.error(f"Ill formatted topic {msg.topic!r};"
                      f"{type(e)}: {e}")
            return

        self._message_callback(module, topic, packet.get('message'))



# Copyright (c) 2021 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Provide a value history in Memory
"""
import logging

from pi_device_net._interfaces import IArchiver

LOG = logging.getLogger(__name__)


class ArchiveMemory(IArchiver):
    """ Manage multiple readings in memory.  Uses a circular buffer to
    minimize re-allocations.
    """

    def __init__(self, size):
        """ Initialize the memory archiver.

        Parameters
        ----------
        size : Int
            The number of samples to hold in memory
        """
        self._buffer = [None] * size
        self._size = size
        self._insert = 0
        self._running = False

    def run(self):
        self._running = True

    def stop(self):
        self._running = False

    def put(self, value):
        """ Add a value to the memory archive, looping through the static
        history buffer.
        """
        if not self._running:
            raise RuntimeError("Memory archive is not running.")

        self._buffer[self._insert] = value
        self._insert = (self._insert + 1) % self._size

    @property
    def history(self, size=None):
        """ Return the history with an optional size to return.

        Slices the buffer to give a coherent result, given the circular
        nature of the buffer.

        Parameters
        ----------
        size : Int
            The number of values to return. If specified larger than the
            buffer size, will be clamped to the buffer size.

        Returns
        -------
        tuple(Number)
            A list of size `size` of the most recent values.  If enough
            samples haven't been put in to fill this size, the return may
            include trailing `None` values.
        """
        size = size or self._size
        # Clamp size to buffer size
        size = min(size, self._size)

        # Determine the sub-areas of the buffer to return...
        # ... insert point towards up to the size or buffer end
        block_a = min(self._size, self._insert + size)
        # ... start of buffer filling out the request size
        block_b = size - (block_a - self._insert)
        # Slice and dice the return!  Cast to tuple to make it immutable.
        return tuple(self._buffer[self._insert:block_a]
                     + self._buffer[0:block_b])



# Copyright (c) 2021 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Useful constants and enumerations, used by PiDeviceNet
"""
from dataclasses import dataclass
from enum import Enum


class ModuleState(Enum):
    """ Module state
    """
    DISCOVERY = 'discovery'
    RUNNING = 'running'
    STOPPED = 'stopped'


class Command(Enum):
    """ Controller/Module messaging commands
    """
    DISCOVER = 'discover'
    FOUND = 'found'
    CONNECT = 'connect'
    MODULE = 'module'
    CONTROLLER = 'controller'
    ATTACH = 'attach'
    DETACH = 'detach'
    PING = 'ping'


# Multicast addressing:  https://tools.ietf.org/html/rfc2365
# IPv4 Organization Local Scope
# the space from which an organization should allocate sub-
# ranges when defining scopes for private use
MULTICAST_GROUP = '239.192.0.1'

# Port; arbitrary from the list of Dynamic ports
# https://en.wikipedia.org/wiki/List_of_TCP_and_UDP_port_numbers#Dynamic,_private_or_ephemeral_ports
MULTICAST_PORT = 49152

# Maximum packet size for discovery messages (just for the sockets used
# during discovery; actual connected messaging is not under an explicit limit)
DISCOVERY_PACKET_SIZE = 1024

SAMPLE_RATE = 1000


# === Various message data classes for visibility (here) and control

@dataclass
class ControllerDiscover:
    system: str
    host: str
    command: str = Command.DISCOVER.value


@dataclass
class ControllerConnectReply:
    system: str
    broker: str
    id: str
    topic: str
    command: str = Command.CONNECT.value


@dataclass
class ControllerPublish:
    system: str
    message: any
    command: str = Command.CONTROLLER.value


@dataclass
class ModuleFound:
    module: str
    id: str
    host: str
    command: str = Command.FOUND.value


@dataclass
class ModulePublish:
    system: str
    module: str
    id: str
    message: any
    command: str = Command.MODULE.value


@dataclass
class ModuleAttachMsg:
    label: str
    entry_page: str
    pages: list
    command: str = Command.ATTACH.value


@dataclass
class ModuleDetachMsg:
    command: str = Command.DETACH.value


@dataclass
class ControllerDetachMsg:
    command: str = Command.DETACH.value


@dataclass
class ControllerPingMsg:
    module: str
    command: str = Command.PING.value


@dataclass
class ModulePingMsg:
    command: str = Command.PING.value


@dataclass
class ModuleMessageMsg:
    module: str
    command: str
    data: dict



# Copyright (c) 2021 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Rudimentary online statistics, providing an approximate mean value
    of the data stream.
"""
from pi_device_net._interfaces import IStatistics


class StatisticsFading(IStatistics):
    def __init__(self, window_size, oldest_weight):
        """ Online statistics, where the older samples have less weight than
        newer samples. No samples are stored, only three internal values.

        Parameters
        ----------
        window_size : Integer
            How many samples we are representing
        oldest_weight : Float
            The relative weight of the oldest sample.  Should be a fairly
            small value.
        """
        self.reset()
        self._init(window_size, oldest_weight)

    def _init(self, window_size, oldest_weight):
        self.alpha = pow(oldest_weight, 1.0 / window_size)

    def reset(self):
        self.sum = 0.0
        self.sum2 = 0.0
        self.increment = 0.0

    def set(self, alpha, sum, sum2, increment):
        self.alpha = alpha
        self.sum = sum
        self.sum2 = sum2
        self.increment = increment

    def update(self, sample):
        self.sum = (self.alpha * self.sum) + sample
        self.increment = (self.alpha * self.increment) + 1.0

    def mean(self):
        # If we haven't gotten enough data, we return a mean of zero (to
        # avoid floating point explosions)
        if self.increment > 0.01:
            return self.sum / self.increment
        else:
            return 0.0
